/*
 *  typemap.cpp
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 *
 *  Windows platform specific code copyright (C) Lauri Pesonen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include "sysdeps.h"
#include "macos_util.h"
#include "prefs.h"
#include "typemap.h"

ext2type default_e2t_translation[] = {
	{".Z", FOURCC('Z','I','V','M'), FOURCC('L','Z','I','V')},
	{".gz", FOURCC('G','z','i','p'), FOURCC('G','z','i','p')},
	{".hqx", FOURCC('T','E','X','T'), FOURCC('S','I','T','x')},
	{".pdf", FOURCC('P','D','F',' '), FOURCC('C','A','R','O')},
	{".ps", FOURCC('T','E','X','T'), FOURCC('t','t','x','t')},
	{".sit", FOURCC('S','I','T','!'), FOURCC('S','I','T','x')},
	{".tar", FOURCC('T','A','R','F'), FOURCC('T','A','R',' ')},
	{".uu", FOURCC('T','E','X','T'), FOURCC('S','I','T','x')},
	{".uue", FOURCC('T','E','X','T'), FOURCC('S','I','T','x')},
	{".zip", FOURCC('Z','I','P',' '), FOURCC('Z','I','P',' ')},
	{".8svx", FOURCC('8','S','V','X'), FOURCC('S','N','D','M')},
	{".aifc", FOURCC('A','I','F','C'), FOURCC('T','V','O','D')},
	{".aiff", FOURCC('A','I','F','F'), FOURCC('T','V','O','D')},
	{".au", FOURCC('U','L','A','W'), FOURCC('T','V','O','D')},
	{".mid", FOURCC('M','I','D','I'), FOURCC('T','V','O','D')},
	{".midi", FOURCC('M','I','D','I'), FOURCC('T','V','O','D')},
	{".mp2", FOURCC('M','P','G',' '), FOURCC('T','V','O','D')},
	{".mp3", FOURCC('M','P','G',' '), FOURCC('T','V','O','D')},
	{".wav", FOURCC('W','A','V','E'), FOURCC('T','V','O','D')},
	{".bmp", FOURCC('B','M','P','f'), FOURCC('o','g','l','e')},
	{".gif", FOURCC('G','I','F','f'), FOURCC('o','g','l','e')},
	{".lbm", FOURCC('I','L','B','M'), FOURCC('G','K','O','N')},
	{".ilbm", FOURCC('I','L','B','M'), FOURCC('G','K','O','N')},
	{".jpg", FOURCC('J','P','E','G'), FOURCC('o','g','l','e')},
	{".jpeg", FOURCC('J','P','E','G'), FOURCC('o','g','l','e')},
	{".pict", FOURCC('P','I','C','T'), FOURCC('o','g','l','e')},
	{".png", FOURCC('P','N','G','f'), FOURCC('o','g','l','e')},
	{".sgi", FOURCC('.','S','G','I'), FOURCC('o','g','l','e')},
	{".tga", FOURCC('T','P','I','C'), FOURCC('o','g','l','e')},
	{".tif", FOURCC('T','I','F','F'), FOURCC('o','g','l','e')},
	{".tiff", FOURCC('T','I','F','F'), FOURCC('o','g','l','e')},
	{".htm", FOURCC('T','E','X','T'), FOURCC('M','O','S','S')},
	{".html", FOURCC('T','E','X','T'), FOURCC('M','O','S','S')},
	{".txt", FOURCC('T','E','X','T'), FOURCC('t','t','x','t')},
	{".rtf", FOURCC('T','E','X','T'), FOURCC('M','S','W','D')},
	{".c", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".cc", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".cp", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".cpp", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".cxx", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".h", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".hh", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".hpp", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".hxx", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".s", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".i", FOURCC('T','E','X','T'), FOURCC('C','W','I','E')},
	{".mpg", FOURCC('M','P','E','G'), FOURCC('T','V','O','D')},
	{".mpeg", FOURCC('M','P','E','G'), FOURCC('T','V','O','D')},
	{".mov", FOURCC('M','o','o','V'), FOURCC('T','V','O','D')},
	{".fli", FOURCC('F','L','I',' '), FOURCC('T','V','O','D')},
	{".avi", FOURCC('V','f','W',' '), FOURCC('T','V','O','D')},
	{".qxd", FOURCC('X','D','O','C'), FOURCC('X','P','R','3')},
	{".hfv", FOURCC('D','D','i','m'), FOURCC('d','d','s','k')},
	{".dsk", FOURCC('D','D','i','m'), FOURCC('d','d','s','k')},
	{".img", FOURCC('r','o','h','d'), FOURCC('d','d','s','k')},
	{0, 0, 0}	// End marker
};

ext2type *e2t_translation = 0;
static bool e2t_allocated = false;

void get_typemap_file_name( HINSTANCE h, char *path )
{
	const char *user_path = PrefsFindString("typemapfile");
	if(user_path) {
		strcpy( path, user_path );
	} else {
		GetModuleFileName( (HMODULE)h, path, _MAX_PATH );
		char *p = strrchr( path, '\\' );
		if(p) {
			*(++p) = 0;
		} else {
			*path = 0;
		}
		strcat( path, "BasiliskII.ftm" );
	}
}

void final_type_map( void )
{
	if(e2t_allocated) {
		free(e2t_translation);
		e2t_translation = 0;
	}
}

void myquote( unsigned char *buf, BOOL encode )
{
	int i, len = strlen((char *)buf), tmplen;
	unsigned char tmp[100];

	if(encode) {
		for( i=len-1; i>=0; i-- ) {
			if(buf[i] < 32 || buf[i] > 127 || buf[i] == '\\') {
				wsprintf( (char *)tmp, "\\%03d", (int)buf[i] );
				tmplen = strlen((char *)tmp); // always 4...
				memmove( &buf[i+tmplen], &buf[i+1], strlen((char *)&buf[i+1])+1 );
				memmove( (char *)&buf[i], (char *)tmp, tmplen );
			}
		}
	} else {
		for( i=len-1; i>=0; i-- ) {
			// "\006..."
			if(buf[i] == '\\' && strlen((char *)&buf[i]) >= 4) {
				buf[i] = (unsigned char)atoi((char *)&buf[i+1]);
				memmove( &buf[i+1], &buf[i+4], strlen((char *)&buf[i+4])+1 );
			}
		}
	}
}

// hqx,,TEXT,SITx,,x,"Default translation"
static bool parse_line( unsigned char *s, int i )
{
	int c = 1;

	e2t_translation[i].ext[0] = '.';
	while( *s >= ' ' && *s != ',' && c < 14 ) {
		e2t_translation[i].ext[c++] = *s++;
	}
	e2t_translation[i].ext[c] = 0;

	if(*s != ',') return false;
	s++;

	if(*s == 'x') s++;

	if(*s != ',') return false;
	s++;

	if(s[4] != ',') return false;
	e2t_translation[i].type =
			((unsigned long)s[0] << 24) |
			((unsigned long)s[1] << 16) |
			((unsigned long)s[2] <<  8) |
			((unsigned long)s[3] <<  0);
	s += 5;

	if(s[4] != ',') return false;
	e2t_translation[i].creator =
			((unsigned long)s[0] << 24) |
			((unsigned long)s[1] << 16) |
			((unsigned long)s[2] <<  8) |
			((unsigned long)s[3] <<  0);

	return true;
}

void init_type_map( HINSTANCE h )
{
	char path[_MAX_PATH];
	get_typemap_file_name( h, path );

	int count = 0;
	FILE *f = fopen(path,"r");

  if(f) {
		e2t_allocated = true;
		int alloc_items = 100;
		int alloc_sz = alloc_items * sizeof(struct ext2type);
		e2t_translation = (struct ext2type *)malloc(alloc_sz);

		if(e2t_translation) {
			char line[256];
			char *linebuf = (char *)malloc( 1000 );
			while(fgets(line, 255, f)) {
				strcpy( linebuf, line );
				myquote((unsigned char *)linebuf,FALSE);
				if(count >= alloc_items) {
					alloc_items += 10;
					alloc_sz = alloc_items * sizeof(struct ext2type);
					e2t_translation = (struct ext2type *)realloc(e2t_translation,alloc_sz);
					if(!e2t_translation) {
						count = 0;
						break;
					}
				}
				if(parse_line( (unsigned char *)linebuf, count )) count++;
			}
			free(linebuf);
		}
		fclose(f);
	}
	if(count == 0) {
		e2t_allocated = false;
		if(e2t_translation) free(e2t_translation);
		e2t_translation = default_e2t_translation;
	}
}
